import logging.config
import os

LOG_LEVEL = os.getenv('LOG_LEVEL', 'WARNING')
LVL_MAP = dict(w='WARNING', d='DEBUG', i='INFO', e='ERROR')

def make_cfg(level=None):
    lvl = LOG_LEVEL
    if level is not None:
        lvl = LVL_MAP.get(level.lower(), LOG_LEVEL)

    
    return {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'verbose': {
            'format': '%(levelname)s %(asctime)s - %(module)s: %(message)s'
        },
        'simple': {
            'format': '%(levelname)s %(message)s'
        },
    },
    'handlers': {
        'console': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'verbose'
        },
        'log-file': {
            'level': 'DEBUG',
            'class': 'logging.StreamHandler',
            'formatter': 'verbose'
        }
    },
    'loggers': {
        'pylunch': {'handlers': ['console'], 'level': lvl},
    }
}


def load(level: str = None):
    cfg = make_cfg(level)
    logging.config.dictConfig(cfg)



import collections
from typing import MutableMapping, Union, Any
from pathlib import Path
import tempfile
import logging
import yaml
import os

log = logging.getLogger(__name__)
CACHE_DIR = Path(tempfile.gettempdir()) / 'pylunch'


def load_yaml(file: Union[Path, str]) -> MutableMapping[str, Any]:
    file = Path(file)

    if not file.exists():
        log.warning(f"[LOAD] Config file not exists: {file}")
        return {}
    with file.open("r") as fp:
        return yaml.safe_load(fp)


def save_yaml(file: Union[Path, str], content: MutableMapping):
    file = Path(file)

    if not file.parent.exists():
        log.warning(f"[SAFE] Unnable to safe config file (directory not exists): {file}")
        return
    with file.open("w") as fp:
        yaml.safe_dump(content, fp)


class YamlLoader:
    def __init__(self, base_dir: Union[str, Path], file: Union[str, Path]):
        self.base_dir = Path(base_dir)
        self.file = Path(file)

    @property
    def full_path(self) -> Path:
        return self.real_path(self.file)

    def load(self) -> MutableMapping:
        file = self.real_path(self.file)
        content = load_yaml(file)
        return content

    def save(self, data: MutableMapping):
        file = self.real_path(self.file)
        log.info(f"[SAVE] Saving content ({file}): {data}")
        save_yaml(file, data)

    def real_path(self, path: Union[str, Path] = None) -> Path:
        path = path if path is not None else self.file
        if path.is_absolute():
            return path
        return self.base_dir / path


class AppConfig(collections.MutableMapping):
    def __init__(self, **kwargs):
        self._config = {**kwargs}

    @property
    def config(self) -> MutableMapping[str, Any]:
        return self._config

    def __getitem__(self, k):
        return self.config.get(k)

    def __setitem__(self, k, v):
        self.config[k] = v

    def __delitem__(self, k):
        del self.config[k]

    def __iter__(self):
        return iter(self.config)

    def __len__(self):
        return len(self.config)

    @property
    def restaurants(self) -> Path:
        return Path(self.config.get('restaurants', './restaurants.yml'))

    @property
    def use_cache(self) -> bool:
        return not self.config.get('no_cache', False)

    @property
    def cache_dir(self) -> Path:
        return Path(self.config.get('cache_dir', os.getenv('PYLUNCH_CACHE_DIR', CACHE_DIR)))

    @property
    def format(self) -> str:
        return self.config.get('format', 'text')

    @property
    def zomato_key(self) -> str:
        return self.config.get('zomato_key', None)

    @property
    def telegram_token(self) -> str:
        return self.config.get('telegram_token', None)



class PyLunchError(Exception):
    def __init__(self, message):
        super().__init__(message)
        self.message = message


class PyLunchApiError(PyLunchError):
    def __init__(self, message, code=400):
        super().__init__(message)
        self.code = code

    def to_json(self):
        return dict(message=self.message, status=self.code)


class UnnableToLoadContent(PyLunchApiError):
    def __init__(self, name, code=400):
        super().__init__(f"Cannot parse the menu for: {name}")
        self.name = name

    def to_json(self):
        return dict(message=self.message, name=self.name, status=400)


import datetime
import logging
import yaml
import sys
from pathlib import Path
from typing import List, Tuple, Mapping
from shutil import copyfile

import click

from pylunch import log_config, lunch, __version__, config, utils

log = logging.getLogger(__name__)

base_dir = Path(__file__).parent.parent
RESOURCES = base_dir / 'resources'
APP_NAME = 'PyLunch'
CONFIG_DIR = click.get_app_dir(APP_NAME.lower())

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])


class CliApplication:
    def __init__(self, config_dir=None):
        self.service: lunch.LunchService = None
        config_dir = config_dir if config_dir is not None else CONFIG_DIR
        self.config_loader = config.YamlLoader(config_dir, 'config.yaml')
        self.restaurants_loader = config.YamlLoader(config_dir, 'restaurants.yaml')

    def init(self, no_zomato=False, **kwargs) -> 'CliApplication':
        if not self.config_loader.base_dir.exists():
            self._first_run()
        cfg_dict = {**self.config_loader.load(), **kwargs}
        if no_zomato and 'zomato_key' in cfg_dict:
            del cfg_dict['zomato_key']
        cfg = config.AppConfig(**cfg_dict)
        loaded = self.restaurants_loader.load() or dict(restaurants={})
        unwrapped = loaded.get('restaurants') or loaded
        upsdated_str = loaded.get('updated')
        updated = datetime.datetime.fromisoformat(upsdated_str) if upsdated_str is not None else None
        ent = lunch.Entities(unwrapped, updated=updated)

        self.service = lunch.LunchService(cfg, ent)
        return self

    def _first_run(self):
        log.info(f"First run detected, crearing config folder: {self.config_loader.base_dir}")
        self.config_loader.base_dir.mkdir(parents=True)
        self.config_loader.save(data=dict(restaurants='./restaurants.yaml'))
        self.restaurants_loader.save(data={})

    def save_restaurants(self):
        log.info("Saving restaurants")
        self.restaurants_loader.save(self.service.instances.to_dict())

    def select_instances(self, selectors, fuzzy=False, tags=False, with_disabled=True) -> List[lunch.LunchEntity]:
        return self.service.instances.select(selectors, fuzzy=fuzzy, tags=tags, with_disabled=with_disabled)


pass_app = click.make_pass_decorator(CliApplication)


@click.group(help=f'{APP_NAME} CLI tool', context_settings=CONTEXT_SETTINGS)
@click.version_option(version=__version__)
@click.option('-L', '--log-level', help=f'Set log level (d|i|w|e) - default=w', default=None)
@click.option('-C', '--no-cache', help=f'Disable cache', is_flag=True, default=False)
@click.option('-c', '--config-dir', help=f'Location to the configuration directory', default=None)
@click.option('-F', '--format', help='Set output format', default=None)
@click.option('--no-zomato', help='Disable zomato even if enabled', default=False, is_flag=True)
@click.pass_context
def main_cli(ctx=None, log_level=None, format=None, no_cache=False, config_dir=None, no_zomato=False, **kwargs):
    log_config.load(log_level)
    app = CliApplication(config_dir=config_dir)
    ctx.obj = app.init(no_cache=no_cache, format=format, log_level=log_level, no_zomato=no_zomato)


@main_cli.command(name='ls', help='List all available restaurants')
@click.option('-l', '--limit', help="Limits number of restaurants to be shown", required=False, default=None)
@pass_app
def cli_list(app: CliApplication, limit=None):
    print(app.service.to_string())


@main_cli.command(name='version', help='Show version')
@pass_app
def cli_version(app: CliApplication):
    print(f"Version: {__version__}")


@main_cli.command(name='menu', help='Get menu for a restaurant')
@click.argument('selectors', nargs=-1)
@click.option("-t", "--tags", help="Search by tags", default=False, is_flag=True)
@click.option("-U", "--update-cache", '--update', help="Update cache entry", default=False, is_flag=True)
@click.option("--cut-before", help="Remove content before the substring", default=None)
@click.option("--cut-after", help="Remove content after the substring", default=None)
@click.option("--no-filters", help="Do not apply filters", default=False, is_flag=True)
@click.option("-F", "--full", help="Show full output - do not apply day filter", default=False, is_flag=True)
@click.option("-Q", "--with-fails", help="Show also fails at the end", default=False, is_flag=True)
@pass_app
def cli_menu(app: CliApplication, selectors: Tuple[str], tags=False, update_cache=False, **kwargs):
    instances = app.select_instances(selectors, tags=tags, with_disabled=False)
    if update_cache:
        cleared = app.service.cache.clear(instances)
    print_instances(app.service, instances, **kwargs)


@main_cli.command(name='roll', help='Get random menu for a restaurant')
@click.argument('selectors', nargs=-1)
@click.option("-t", "--tags", help="Search by tags", default=False, is_flag=True)
@click.option("-l", "--limit", help="Limit number of restaurants (default: 1)", default=1)
@pass_app
def cli_roll(app: CliApplication, selectors: Tuple[str], tags=False, limit=1, **kwargs):
    instances = app.select_instances(selectors, tags=tags, with_disabled=False)
    if instances:
        import random
        selected = random.choices(instances, k=limit)
        print_instances(app.service, selected, **kwargs)
    else:
        print("No instance found :-(")


@main_cli.command(name='info', help='Get info for the restaurant')
@click.argument('selectors', nargs=-1)
@click.option("-t", "--tags", help="Search by tags", default=False, is_flag=True)
@pass_app
def cli_info(app: CliApplication, selectors=None, tags=False):
    instances = app.select_instances(selectors, tags=tags)
    print_instances(app.service, instances, transform=str)


@main_cli.command(name='import', help='Import restaurants')
@click.argument('restaurants', nargs=-1)
@click.option("-O", "--override", help="Overide the restaurant if exists", default=False, is_flag=True)
@pass_app
def cli_import(app: CliApplication, restaurants=None, override=False):
    if not restaurants:
        print("Importing from the stdin. [EOF for END]")
        app.service.import_string(sys.stdin.read(), override=override)
    else:
        for rest_file in restaurants:
            print(f"Importing restaurant: {rest_file}")
            app.service.import_file(rest_file, override=override)
    app.save_restaurants()


@main_cli.command(name='export', help='Export restaurants')
@click.option("-f", "--file", help="Export to file", default=None)
@pass_app
def cli_export(app: CliApplication, file=None):
    if file is None:
        if app.restaurants_loader.full_path.exists():
            content = app.restaurants_loader.full_path.read_text('utf-8')
            print(content)
        else:
            print(f"Error: Restaurants file not exists: {app.restaurants_loader.full_path}")

    else:
        # not a hack :-)
        copyfile(src=str(app.restaurants_loader.full_path), dst=file)


@main_cli.command(name='add', help='Adds a new restaurant')
@click.option("-n", "--name", help="Restaurant name", default=False)
@click.option("-d", "--display-name", help="Restaurant name", default=False)
@click.option("-u", "--url", help="Restaurant url", default=False)
@click.option("-s", "--selector", help="Restaurant css selector", default=False)
@click.option("-t", "--tags", help="Restaurant tags", default=False, multiple=True)
@click.option("-p", "--param", help="Additional param", default=False, multiple=True)
@click.option("-O", "--override", help="Overide the restaurant if exists", default=False, is_flag=True)
@pass_app
def cli_add(app: CliApplication, name, display_name, url, selector, tags, params, override=False):
    tags = list(tags) if tags else []
    config = dict(name=name, url=url, display_name=display_name, tags=tags, selector=selector)
    if params:
        config.update(**_params_dict(params), override=override)
    app.service.instances.register(**config)
    app.save_restaurants()


@main_cli.command(name='rm', help='Removes the restaurant')
@click.argument('selectors', nargs=-1)
@click.option("-f", "--fuzzy", help="Fuzzy search the name", default=False, is_flag=True)
@click.option("-t", "--tags", help="Search by tags", default=False, is_flag=True)
@pass_app
def cli_rm(app: CliApplication, selectors: Tuple[str], fuzzy=False, tags=False):
    instances = app.select_instances(selectors, fuzzy=fuzzy, tags=tags)
    for instance in instances:
        print(f"Removing: {instance.name}")
        del app.service.instances[instance.name]
    app.save_restaurants()


@main_cli.command(name='enable', help='Enables the restaurants')
@click.argument('selectors', nargs=-1)
@click.option("-f", "--fuzzy", help="Fuzzy search the name", default=False, is_flag=True)
@click.option("-t", "--tags", help="Search by tags", default=False, is_flag=True)
@pass_app
def cli_enable(app: CliApplication, selectors: Tuple[str], fuzzy=False, tags=False):
    instances = app.select_instances(selectors, fuzzy=fuzzy, tags=tags)
    for instance in instances:
        if 'disabled' in instance.keys():
            print(f"Enabling instance {instance.name}: {instance}")
            del app.service.instances[instance.name]['disabled']
    app.save_restaurants()


@main_cli.command(name='disable', help='Disables the restaurants')
@click.argument('selectors', nargs=-1)
@click.option("-f", "--fuzzy", help="Fuzzy search the name", default=False, is_flag=True)
@click.option("-t", "--tags", help="Search by tags", default=False, is_flag=True)
@pass_app
def cli_disable(app: CliApplication, selectors: Tuple[str], fuzzy=False, tags=False):
    instances = app.select_instances(selectors, fuzzy=fuzzy, tags=tags)
    for instance in instances:
        print(f"Disabling instance {instance.name}: {instance}")
        app.service.instances[instance.name]['disabled'] = True
    app.save_restaurants()


@main_cli.command(name='edit', help='Edits restaurants DB file')
@pass_app
def cli_edit_restaurants(app: CliApplication):
    cfg = app.restaurants_loader.load()
    content = yaml_edit(cfg)
    if content is None:
        print("No change - not saving")
    else:
        app.restaurants_loader.save(content)


@main_cli.command(name='config', help='Shows the current configuration')
@pass_app
def cli_config(app: CliApplication):
    import yaml
    cfg = app.service.config.config
    print(yaml.safe_dump(dict(Config=cfg)))


@main_cli.command(name='cache-clear', help='Clear a current cache for a day')
@click.argument('selectors', nargs=-1)
@click.option("-f", "--fuzzy", help="Fuzzy search the name", default=False, is_flag=True)
@click.option("-t", "--tags", help="Search by tags", default=False, is_flag=True)
@pass_app
def cli_cache_clear(app: CliApplication, selectors: Tuple[str], fuzzy=False, tags=False):
    if not app.service.config.use_cache:
        print("Not using the cache - no action.")
        return

    if not selectors:
        for item in app.service.cache.clear():
            print(f"Clearing: {item}")
    else:
        instances = app.select_instances(selectors, fuzzy=fuzzy, tags=tags)
        for item in app.service.cache.clear(instances):
            print(f"Clearing: {item}")


@main_cli.command(name='cache-content', help='Show the current cache for a day')
@pass_app
def cli_cache_content(app: CliApplication):
    if not app.service.config.use_cache:
        print("Not using the cache - no action.")
        return

    content = app.service.cache.content()
    if not content:
        print("No content in cache for today")
    else:
        for path in content:
            print(path)


@main_cli.command(name='cfg-set', help='Set a config value in the user configuration')
@click.argument('name')
@click.argument('value')
@pass_app
def cli_set_config(app: CliApplication, name, value):
    cfg = app.config_loader.load()
    cfg[name] = value
    app.config_loader.save(cfg)


@main_cli.command(name='cfg-edit', help='Edit a configuration using the editor (Ex: VIM)')
@pass_app
def cli_edit_config(app: CliApplication):
    cfg = app.config_loader.load()
    content = yaml_edit(cfg)
    if content is None:
        print("No change - not saving")
    else:
        app.config_loader.save(content)


@main_cli.command(name='console', help='Start the interactive console (IPython)')
@pass_app
def cli_start_console(app: CliApplication):
    try:
        import IPython
        print("Starting the interactive console using the IPython")
        IPython.embed()
    except ImportError:
        print('\nIPython modeule is not available')


@main_cli.command(name='telegram-bot', help='Start the telegram bot')
@pass_app
def cli_start_console(app: CliApplication):
    try:
        from pylunch.telegram_bot import PyLunchTelegramBot
        print("Starting the telegram bot")
        bot = PyLunchTelegramBot(app.service)
        bot.run()
    except ImportError:
        print('\n Telegram bot is not available modeule is not available')


"""
" Helper tools
"""


def yaml_edit(cfg) -> dict:
    content = click.edit(yaml.safe_dump(cfg))
    if content is not None:
        return yaml.safe_load(content)
    return None


def _params_dict(params: List[str]) -> Mapping:
    result = {}
    for param in params:
        (key, val) = param.split('=')
        result[key] = val


def print_instances(service: lunch.LunchService, instances, transform=None, with_fails=False, **kwargs):
    fails = list() if with_fails else None
    transform = transform if transform is not None else lambda x: resolve_menu(service, x, fails=fails, **kwargs)
    utils.write_instances(instances, transform=transform, writer=print)
    if fails:
        print("\n~~~~~~~~~~~~~~~~  FAILS  ~~~~~~~~~~~~~~~\n")
        for fail in fails:
            print(fail)


def resolve_menu(service: lunch.LunchService, instance: lunch.LunchEntity, fails: list = None, **kwargs):
    result = _generate_menu_header(instance)
    content = service.resolve_text(instance, **kwargs)
    if not content:
        if fails is not None:
            fails.append(instance)
        result += f"No content provided for: {instance.name}"
    else:
        result += content
    return result


def _generate_menu_header(instance):
    name_str = f"{instance.display_name} ({instance.name})"
    tags_str = "Tags: " + (", ".join(instance.tags) if instance.tags else '')
    return utils.generate_nice_header(name_str, instance.url, tags_str)


if __name__ == '__main__':
    main_cli()



import flask
import os
from urllib.error import HTTPError
import logging
import click
import datetime
import requests

from pathlib import Path
from typing import List, Mapping, Optional
from pylunch import config, lunch, utils, __version__, log_config, errors
from werkzeug.security import generate_password_hash, check_password_hash


from flask_jwt_extended import (
    JWTManager, jwt_required, create_access_token,
    jwt_refresh_token_required, create_refresh_token,
    get_jwt_identity, set_access_cookies,
    set_refresh_cookies
)

log = logging.getLogger(__name__)

# Find the correct template folder when running from a different location
base_dir = Path(__file__).parent.parent
RESOURCES = base_dir / 'resources'
INTERNAL = base_dir / 'internal'
APP_NAME = 'PyLunch'
CONFIG_DIR = click.get_app_dir(APP_NAME.lower())
tmpl_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "templates")
static_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "static")


class AdminUsers(utils.CollectionWrapper):
    @property
    def users(self) -> Mapping:
        return self.collection

    def import_users(self, file: str):
        log.info(f"[IMPORT] Importing users from: {file}")
        loaded = utils.load_yaml(file)
        if not loaded:
            return

        for (name, password) in loaded.items():
            log.info(f"[IMPORT] Importing user: {name}")
            self[name] = password

    @classmethod
    def generate_hash(self, password):
        return generate_password_hash(password)

    def check_password(self, name, passwd) -> bool:
        userhash = self.get(name)
        if not userhash:
            log.info(f"[USER] Password check failed: User {name} not exists.")
            return False
        if check_password_hash(userhash, passwd):
            log.info(f"[USER] User password accepted for {name}")
            return True
        log.info(f"[USER] User password rejected for {name} - mismatch")
        return False

    def issue_tokens(self, name) -> dict:
        access_token = create_access_token(identity=name)
        refresh_token = create_refresh_token(identity=name)
        return dict(access_token=access_token, refresh_token=refresh_token, username=name)

    def issue_access_token(self, name) -> dict:
        access_token = create_access_token(identity=name)
        return dict(access_token=access_token, username=name)


def handle_general_error(e):
    code = 500
    if isinstance(e, HTTPError):
        code = e.code
    web_app = WebApplication.get()
    context = web_app.gen_context(code=code, stacktrace=e, message=str(e), flash='')
    return flask.render_template('error.html', **context), code


def handle_api_error(e: errors.PyLunchApiError):
    return flask.jsonify(e.to_json()), e.code


def page_not_found(e: Exception):
    # note that we set the 404 status explicitly
    code = 404
    web_app = WebApplication.get()
    context = web_app.gen_context(code=code, stacktrace=e, message=str(e))
    return flask.render_template('error.html', **context), 404


class WebApplication:
    INSTANCE = None

    @classmethod
    def create_app(cls):
        flask_app = flask.Flask(__name__, template_folder=tmpl_dir, static_folder=static_dir)
        flask_app.config['JWT_SECRET_KEY'] = os.getenv('PYLUNCH_SECRET', 'jwt-secret-string')
        flask_app.config['JWT_TOKEN_LOCATION'] = ('cookies', 'headers')
        flask_app.config['JWT_ACCESS_COOKIE_PATH'] = '/'
        flask_app.config['JWT_REFRESH_COOKIE_PATH'] = '/'
        flask_app.config['JWT_COOKIE_CSRF_PROTECT'] = False

        flask_app.config['JWT_ACCESS_TOKEN_EXPIRES'] = datetime.timedelta(days=7)
        flask_app.config['JWT_REFRESH_TOKEN_EXPIRES'] = datetime.timedelta(days=30)

        JWTManager(flask_app)
        if flask_app.debug:
            flask_app.jinja_env.auto_reload = True
            flask_app.config['TEMPLATES_AUTO_RELOAD'] = True
        else:
            flask_app.register_error_handler(404, page_not_found)
            flask_app.register_error_handler(Exception, handle_general_error)
            flask_app.register_error_handler(errors.PyLunchApiError, handle_api_error)
        return flask_app

    def __init__(self, config_dir=None):
        self._service: Optional[lunch.LunchService] = None
        config_dir = config_dir if config_dir is not None else CONFIG_DIR
        self.config_loader = config.YamlLoader(config_dir, 'config.yaml')
        self.restaurants_loader = config.YamlLoader(config_dir, 'restaurants.yaml')
        self.users = AdminUsers()
        self._timestamp = None
        self._config = None

    @property
    def request(self) -> flask.Request:
        return flask.request

    @property
    def service(self) -> lunch.LunchService:
        if self._service is None or ((self._timestamp + datetime.timedelta(minutes=10)) < datetime.datetime.now()):
            self.reload_restaurants()
        return self._service

    def init(self, **kwargs) -> 'WebApplication':
        log_config.load('i')
        if not self.config_loader.base_dir.exists():
            self._first_run()
        cfg_dict = {**self.config_loader.load(), **kwargs}
        self._config = config.AppConfig(**cfg_dict)
        self.users.import_users(os.getenv('PYLUNCH_USERS', RESOURCES / 'users.yml'))
        return self

    def reload_restaurants(self):
        loaded = self.restaurants_loader.load() or dict(restaurants={})
        unwrapped = loaded.get('restaurants') or loaded
        log.info(f"[INIT] Loaded: {[name for name in unwrapped.keys()]}")
        upsdated_str = loaded.get('updated')
        updated = datetime.datetime.fromisoformat(upsdated_str) if upsdated_str is not None else None
        ent = lunch.Entities(unwrapped, updated=updated)
        self._timestamp = datetime.datetime.now()
        self._service = lunch.LunchService(self._config, ent)

    def _first_run(self):
        log.info(f"First run detected, crearing config folder: {self.config_loader.base_dir}")
        self.config_loader.base_dir.mkdir(parents=True)
        self.config_loader.save(data=dict(restaurants='./restaurants.yaml'))
        self.restaurants_loader.save(data={})

    def save_restaurants(self):
        log.info("Saving restaurants")
        self.restaurants_loader.save(self.service.instances.to_dict())

    def select_instances(self, selectors, fuzzy=False, tags=False, with_disabled=True) -> List[lunch.LunchEntity]:
        return self.service.instances.select(selectors, fuzzy=fuzzy, tags=tags, with_disabled=with_disabled)

    def gen_context(self, **kwargs):
        tags = self.service.instances.all_tags()
        restaurants = self.service.instances.all()
        analytics = self._load_analytics()
        return dict(
            version=__version__,
            all_tags=tags,
            all_restaurants=restaurants,
            analytics=analytics,
            **kwargs)

    def _load_analytics(self):
        analytics_path: Path = INTERNAL / 'analytics.html'
        if not analytics_path.exists():
            log.debug(f"[INIT] Analytics not loaded: {analytics_path}")
            return None

        log.info(f"[INIT] Analytics loaded: {analytics_path}")
        return analytics_path.read_text(encoding='utf-8')

    @classmethod
    def get(cls) -> 'WebApplication':
        if cls.INSTANCE is None:
            cls.INSTANCE = cls(config_dir=CONFIG_DIR)
            cls.INSTANCE.init()
        return cls.INSTANCE

    def parse_request(self):
        rq = flask.request
        args = rq.args
        result = dict(selectors=rq.args.getlist('r'), tags=rq.args.getlist('t'), format=rq.args.get('f', 'h'),
                      roll=args.get('roll'))
        return result

    def select_by_request(self):
        args = self.parse_request()
        tags = args['tags']
        selectors = args['selectors']
        format: str = args['format']
        roll = args['roll']

        def _inner():
            if selectors:
                return self.select_instances(selectors)
            elif tags:
                return self.select_instances(tags, tags=True)
            else:
                return self.select_instances(selectors=None)

        result = _inner()
        return roll_filter(result, roll)


app = WebApplication.create_app()
api = flask.Blueprint('api', __name__)
admin = flask.Blueprint('admin', __name__)


def register_blueprints(app):
    app.register_blueprint(api, url_prefix='/api')
    app.register_blueprint(admin, url_prefix='/admin')


def roll_filter(items, roll):
    if not items:
        return []
    if not roll:
        return items
    import random
    return random.choices(items, k=int(roll))


@app.route('/')
def index():
    web_app = WebApplication.get()
    context = web_app.gen_context()
    return flask.render_template('index.html', **context)


@app.route('/restaurants/<name>')
def restaurant(name):
    web_app: WebApplication = WebApplication.get()
    entity = web_app.service.instances.find_one(name)
    menu = web_app.service.resolve_text(entity)
    context = web_app.gen_context(entity=entity, menu=menu)
    return flask.render_template('restaurant.html', **context)


@app.route('/menu')
def web_async_menu():
    web_app = WebApplication.get()
    context = web_app.gen_context()
    return flask.render_template('menu.html', **context)


@app.route("/fmenu")
def web_fallback_menu():
    web_app = WebApplication.get()
    instances = web_app.select_by_request()
    instances = instances if instances is not None else []
    format = web_app.parse_request()['format']

    if format is not None and format.startswith('t'):
        content = "\n".join(resolve_menu(web_app.service, inst) for inst in instances)
        return flask.Response(content, mimetype='text/plain')
    else:
        menus = [(rest, web_app.service.resolve_text(rest)) for rest in instances if rest]
        context = web_app.gen_context(restaurants=instances, menus=menus)
        return flask.render_template('fmenu.html', **context)


###
# API
###

@api.route("/restaurants")
def route_api_restaurants():
    web_app = WebApplication.get()
    instances = web_app.select_by_request()
    return flask.jsonify({item.name: item.config for item in instances if item})


@api.route("/tags")
def route_api_tags():
    web_app = WebApplication.get()
    tags = web_app.service.instances.all_tags()
    return flask.jsonify(tags)


@api.route("/restaurants/<name>")
def route_api_restaurants_get(name):
    web_app = WebApplication.get()
    instance = web_app.service.instances.find_one(name)
    return flask.jsonify(instance.config)


@api.route("/restaurants/<name>/menu")
def route_api_restaurants_get_menu(name):
    web_app = WebApplication.get()
    instance = web_app.service.instances.find_one(name)
    content = web_app.service.resolve_text(instance)
    if content:
        result = {**instance.config, 'content': content}
        return flask.jsonify(result)
    else:
        return flask.jsonify(errors.UnnableToLoadContent(name).to_json()), 400


@api.route("/restaurants/<name>/cache")
def route_api_restaurants_get_cache(name):
    web_app = WebApplication.get()
    instance = web_app.service.instances.find_one(name)
    paths = web_app.service.cache.paths_for_entity(instance, relative=True)
    return flask.jsonify([str(item) for item in paths])


###
# Admin
###

# Same thing as login here, except we are only setting a new cookie
# for the access token.
@admin.route('/token/refresh', methods=['POST'])
@jwt_refresh_token_required
def admin_refresh():
    # Create the new access token
    current_user = get_jwt_identity()
    web_app = WebApplication.get()
    access_token = web_app.users.issue_access_token(current_user)

    # Set the JWT access cookie in the response
    resp = flask.jsonify({'refresh': True, **access_token})
    set_access_cookies(resp, access_token['access_token'])
    return resp, 200


# Use the set_access_cookie() and set_refresh_cookie() on a response
# object to set the JWTs in the response cookies. You can configure
# the cookie names and other settings via various app.config options
@admin.route('/token/auth', methods=['POST'])
def admin_login():
    log.info(f"[LOGIN] Received data: {flask.request.data}")
    username = flask.request.form.get('username', None)
    password = flask.request.form.get('password', None)
    if not username or not password:
        return flask.jsonify({'message': 'Username or password is missing'}), 401

    web_app = WebApplication.get()
    if not web_app.users.check_password(username, password):
        return flask.jsonify({'error': 'Invalid password'}), 401

    tokens = web_app.users.issue_tokens(username)

    # Set the JWT cookies in the response
    resp = flask.jsonify({'login': True, **tokens})
    set_access_cookies(resp, tokens.get('access_token'))
    set_refresh_cookies(resp, tokens.get('refresh_token'))
    return resp, 200


@admin.route('/login', methods=['GET'])
def admin_login_form():
    web_app = WebApplication.get()
    context = web_app.gen_context()
    return flask.render_template('admin/login.html', **context)


@admin.route('/index', methods=['GET'])
@jwt_required
def admin_index():
    web_app = WebApplication.get()
    user = get_jwt_identity()
    context = web_app.gen_context(user=user)
    return flask.render_template('admin/index.html', **context)


@admin.route('/restaurants/edit', methods=['GET'])
@jwt_required
def admin_edit_restaurants():
    web_app = WebApplication.get()
    user = get_jwt_identity()
    context = web_app.gen_context(user=user)
    return flask.render_template('admin/edit.html', **context)


@admin.route('/token/valid', methods=['POST'])
@jwt_required
def admin_token_valid():
    web_app = WebApplication.get()
    user = get_jwt_identity()
    return flask.jsonify({'valid': True})


@admin.route('/cache-invalidate', methods=['POST'])
@jwt_required
def admin_cache_invalidate():
    web_app = WebApplication.get()
    items = web_app.service.cache.clear()
    return flask.jsonify({'message': "cache updated", "content": items})


@admin.route('/config/restaurants')
@jwt_required
def admin_config_restaurants_get():
    web_app = WebApplication.get()
    return flask.jsonify(dict(content=web_app.restaurants_loader.full_path.read_text(encoding='utf-8')))


@admin.route('/config/restaurants', methods=['POST'])
@jwt_required
def admin_config_restaurants_post():
    web_app = WebApplication.get()

    # little not nice hack
    # we need to have "fresh" data in order not hit the timeout
    web_app.reload_restaurants()
    rq = flask.request
    content = rq.form.get('content', None)
    url = rq.form.get('url', None)
    
    if content:
        web_app.service.import_string(content, override=True)
        
    if url:
        res = requests.get(url)
        if res.ok:
            web_app.service.import_string(res.content, override=True)
    web_app.restaurants_loader.save(web_app.service.instances.to_dict())
    return flask.jsonify(dict(content=web_app.service.instances.to_dict()))


register_blueprints(app)


###
# Helpers
### 

def resolve_menu(service: lunch.LunchService, instance):
    result = _generate_menu_header(instance)
    result += service.resolve_text(instance)
    return result


def _generate_menu_header(instance):
    name_str = f"{instance.display_name} ({instance.name})"
    tags_str = "Tags: " + (", ".join(instance.tags) if instance.tags else '')
    return utils.generate_nice_header(name_str, instance.url, tags_str)



import logging
from typing import List, Optional, Tuple, Any, MutableMapping, Mapping, Union, Type, ValuesView, Dict

import html2text
import requests
import yaml
import json
import datetime
import shutil
import collections
import io
import os
import re
import unidecode
from bs4 import BeautifulSoup, Tag
from requests import Response
from pyzomato import Pyzomato

from fuzzywuzzy import fuzz, process
from pathlib import Path

from .tags_evaluator import TagsEvaluator
from .config import AppConfig

from pdfminer import high_level
import pdfminer.layout

log = logging.getLogger(__name__)

USER_AGENTS = [
    # Chrome
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/60.0.3112.113 Safari/537.36',
    'Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/60.0.3112.90 Safari/537.36',
    'Mozilla/5.0 (Windows NT 5.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/60.0.3112.90 Safari/537.36',
    'Mozilla/5.0 (Windows NT 6.2; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/60.0.3112.90 Safari/537.36',
    'Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/44.0.2403.157 Safari/537.36',
    'Mozilla/5.0 (Windows NT 6.3; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/60.0.3112.113 Safari/537.36',
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/57.0.2987.133 Safari/537.36',
    'Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/57.0.2987.133 Safari/537.36',
    'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/55.0.2883.87 Safari/537.36',
    'Mozilla/5.0 (Windows NT 6.1; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/55.0.2883.87 Safari/537.36',
    # Firefox
    'Mozilla/4.0 (compatible; MSIE 9.0; Windows NT 6.1)',
    'Mozilla/5.0 (Windows NT 6.1; WOW64; Trident/7.0; rv:11.0) like Gecko',
    'Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; WOW64; Trident/5.0)',
    'Mozilla/5.0 (Windows NT 6.1; Trident/7.0; rv:11.0) like Gecko',
    'Mozilla/5.0 (Windows NT 6.2; WOW64; Trident/7.0; rv:11.0) like Gecko',
    'Mozilla/5.0 (Windows NT 10.0; WOW64; Trident/7.0; rv:11.0) like Gecko',
    'Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.0; Trident/5.0)',
    'Mozilla/5.0 (Windows NT 6.3; WOW64; Trident/7.0; rv:11.0) like Gecko',
    'Mozilla/5.0 (compatible; MSIE 9.0; Windows NT 6.1; Trident/5.0)',
    'Mozilla/5.0 (Windows NT 6.1; Win64; x64; Trident/7.0; rv:11.0) like Gecko',
    'Mozilla/5.0 (compatible; MSIE 10.0; Windows NT 6.1; WOW64; Trident/6.0)',
    'Mozilla/5.0 (compatible; MSIE 10.0; Windows NT 6.1; Trident/6.0)',
    'Mozilla/4.0 (compatible; MSIE 8.0; Windows NT 5.1; Trident/4.0; .NET CLR 2.0.50727; .NET CLR 3.0.4506.2152; .NET CLR 3.5.30729)'
]


class LunchEntity(collections.MutableMapping):
    def __init__(self, config: Mapping[str, Any]):
        self._config = {**config}

    def __getitem__(self, k):
        return self._config.get(k)

    def __setitem__(self, k, v):
        self.config[k] = v

    def __delitem__(self, k):
        del self.config[k]

    def __iter__(self):
        return iter(self._config)

    def __len__(self):
        return len(self.config)

    @property
    def config(self) -> MutableMapping['str', Any]:
        return self._config

    @property
    def resolver(self) -> str:
        return self.config.get('resolver', 'default')

    @property
    def name(self) -> str:
        return self.config.get('name')

    @property
    def url(self) -> str:
        return self.config.get('url')

    @property
    def selector(self) -> str:
        return self.config.get('selector')

    @property
    def request_params(self) -> List:
        return self.config.get('request_params', {})

    @property
    def display_name(self) -> str:
        return self.config.get('display_name') or self.name

    @property
    def tags(self) -> List[str]:
        return self.config.get('tags')

    @property
    def disabled(self) -> bool:
        return self.config.get('disabled', False)

    @property
    def days(self) -> List[str]:
        return self.config.get('days')

    @property
    def filters(self) -> str:
        return self.config.get('filters')

    @property
    def resolvers(self) -> List[MutableMapping]:
        return self.config.get('resolvers')

    @property
    def language(self) -> str:
        return self.config.get('language') or 'eng'

    def __str__(self) -> str:
        result = f"\"{self.name}\" -"

        if self.display_name:
            result += f" ({self.display_name})"

        if self.tags:
            result += f" {self.tags}"

        result += f" {self.url}"

        if self.selector:
            result += f" ({self.selector})"
        if self.request_params:
            result += f" - req_params={self.request_params}"

        if self.resolver and self.resolver != 'default':
            result += f' resolver={self.resolver}'
        if self.filters:
            result += f' filters={self.filters}'
        return result

    def __repr__(self) -> str:
        return str(self.config)

    @classmethod
    def clone(cls, origin: 'LunchEntity'):
        return cls(origin.config)


class ResolverConfig(collections.MutableMapping):
    def __init__(self, config: Mapping[str, Any], entity: LunchEntity = None, content=None):
        self._config = {**config}
        self._entity = entity
        self._content = content

    def __getitem__(self, k):
        return self._config.get(k)

    def __setitem__(self, k, v):
        self.config[k] = v

    def __delitem__(self, k):
        del self.config[k]

    def __iter__(self):
        return iter(self._config)

    def __len__(self):
        return len(self.config)

    @property
    def config(self) -> MutableMapping['str', Any]:
        return self._config

    @property
    def entity(self) -> LunchEntity:
        return self._entity

    @property
    def name(self) -> str:
        return self.resolver

    @property
    def resolver(self) -> str:
        return self.config.get('resolver')

    @property
    def request_params(self) -> dict:
        return self.config.get('request_params', {})

    @property
    def url(self) -> str:
        return self.config.get('url')

    @property
    def selector(self) -> str:
        return self.config.get('selector') or self.entity.selector

    @property
    def content(self) -> str:
        return self._content if self._content is not None else self.config.get('content')

    @property
    def text(self) -> bool:
        text = self.config.get('text', 'false')
        from distutils import util
        return util.strtobool(text)

    @property
    def allow_cache(self) -> bool:
        return not self.no_cache

    @property
    def no_cache(self) -> bool:
        no_cache = self.config.get('no_cache', 'false')
        from distutils import util
        return util.strtobool(no_cache)


######
# Resolvers
######


class AbstractResolver:
    CACHE_EXT = 'txt'
    CACHE_SUFFIX = None
    CACHE_DISABLED = False

    def __init__(self, service: 'LunchService', config: ResolverConfig):
        log.debug(f"[RESOLV] Instance of {self.__class__.__name__} with {config}")
        self._service = service
        self._config = config

    @property
    def config(self) -> ResolverConfig:
        return self._config

    @property
    def entity(self) -> LunchEntity:
        return self.config.entity

    @property
    def service(self) -> 'LunchService':
        return self._service

    def resolve(self, day=None, **kwargs) -> Any:
        cls = self.__class__
        log.info(f"[RESOLV] Resolving {self.entity.name} using the {cls.__name__}.")
        try:
            suffix = cls.CACHE_SUFFIX or cls.__name__
            allow_cache = self.config.allow_cache and not cls.CACHE_DISABLED
            log.debug("[RESOLV] Cache state: {}")
            if allow_cache:
                return self.service.cache.wrap(entity=self.entity, func=self._resolve, day=day, ext=cls.CACHE_EXT,
                                               suffix=suffix)

            return self._resolve(day=day, **kwargs)
        except Exception as ex:
            log.error(f"[RESOLV] Resolved error {cls.__name__}: {ex}", exc_info=True)
            return None

    def resolve_text(self, **kwargs) -> str:
        content = self.resolve(**kwargs)
        return None if not content else str(content)

    def _resolve(**kwargs) -> Any:
        """ This method should be overriden - abstract method
        """
        return None


class ResolverChain(AbstractResolver):
    CACHE_EXT = 'txt'
    CACHE_SUFFIX = 'chain'

    @property
    def chain(self) -> List[MutableMapping]:
        return self.entity.resolvers

    def _resolve(self, **kwargs) -> Any:
        content = None
        log.info(f"[CHAIN] Resolving chain for {self.entity.name}")
        for res_cfg in self.chain:
            config = ResolverConfig(config=res_cfg, entity=self.entity, content=content)
            resolved = self._resolve_one(config=config, content=content, **kwargs)

            if not resolved:
                log.warning(f"[CHAIN] Resolver {config.name} for {self.entity.name}: no content")
            else:
                log.info(f"[CHAIN] Resolver {config.name} for {self.entity.name}: {resolved}")
            content = resolved
        return content

    def _resolve_one(self, config, content, **kwargs):
        resolver = self.service.resolvers.get(config.name)
        if not resolver:
            log.warning(f"[CHAIN] Resolver {config.name} for {self.entity.name} was not found, skipping.")
            return None
        log.info(f"[CHAIN] Using the resolver {config.name} for entity: {self.entity.name}: {resolver.__name__}")
        instance: AbstractResolver = resolver(service=self.service, config=config)
        if config.text:
            resolved = instance.resolve(**kwargs)
        else:
            resolved = instance.resolve_text(**kwargs)
        return resolved


class RequestResolver(AbstractResolver):
    CACHE_EXT = 'dat'
    CACHE_SUFFIX = 'raw-request'

    @classmethod
    def random_useragent(cls):
        import random
        return random.choice(USER_AGENTS)

    @property
    def request_url(self) -> str:
        return self.config.content or self.config.url or self.entity.url

    def get_request_params(self) -> dict:
        headers = {'User-Agent': self.__class__.random_useragent()}
        params = dict()
        if self.config.request_params:
            params.update(self.config.request_params)
            if 'headers' in params:
                params['headers'].update(headers)
        else:
            params['headers'] = headers
        return params

    def _resolve(self, **kwargs) -> Optional[requests.Response]:
        try:
            params = self.get_request_params()
            response = requests.get(self.request_url, **(params))
        except Exception as ex:
            log.error(f"Request error: {ex}")
            return None
        if not response.ok:
            log.warning(f"[LUNCH] Error[{response.status_code}] ({self.entity.name}): {response.content}")
            print(f"Error[{response.status_code}] {self.entity.name}: ", response.content)
        else:
            log.debug(f"[RES] Response [{response.status_code}] {self.entity.name}: {response.content}")
        return response

    def resolve_text(self, **kwargs) -> Optional[str]:
        res = self.resolve(**kwargs)
        if res.ok:
            return res.content.decode('utf-8')
        return None


class HtmlResolver(RequestResolver):
    CACHE_EXT = 'html'
    CACHE_SUFFIX = 'html'

    def _parse_response(self, response: Response) -> List[Tag]:
        soap = BeautifulSoup(response.content, "lxml")
        sub = soap.select(self.config.selector) if self.config.selector else soap
        log.debug(f"[LUNCH] Parsed[{self.entity.name}]: {sub}")
        return sub

    def resolve_text(self, **kwargs) -> Optional[str]:
        html_string = self.resolve(**kwargs)
        if html_string is None:
            return None
        return to_text(html_string)

    def _resolve(self, **kwargs) -> Optional[str]:
        response = super()._resolve(**kwargs)
        if response is None:
            return None
        parsed = self._parse_response(response=response)
        content = self.to_string(parsed)
        if not content:
            log.warning(f"[HTML] Content is empty for {self.entity.name} - {self.config.url} ({self.config.selector})")
            return None
        else:
            log.debug(f"[HTML] Extracted content {self.entity.name}: {content}")
        return content

    @classmethod
    def to_string(cls, parsed) -> str:
        if isinstance(parsed, list):
            items = [str(item) for item in parsed]
            return "".join(items)
        else:
            return str(parsed)


class AbstractHtmlResolver(AbstractResolver):
    CACHE_EXT = 'html'
    CACHE_SUFFIX = 'html'

    def resolve_text(self, **kwargs) -> Optional[str]:
        html_string = self.resolve(**kwargs)
        if html_string is None:
            return None
        return to_text(html_string)


class HtmlTagsSelectorResolver(AbstractHtmlResolver):
    CACHE_DISABLED = True

    def _resolve(self, **kwargs) -> List[Tag]:
        response: requests.Response = self.config.content

        soap = BeautifulSoup(response.content, "lxml")
        tags = soap.select(self.entity.selector) if self.entity.selector else soap
        log.debug(f"[LUNCH] Parsed[{self.entity.name}]: {tags}")
        return tags


class ToStringResolver(AbstractResolver):
    def _resolve(self, **kwargs) -> Optional[str]:
        string = self.config.content
        content = self.to_string(string)
        if not content:
            log.warning(f"[STR] Content is empty for {self.entity.name}")
            return None
        else:
            log.debug(f"[STR] Extracted content {self.entity.name}: {content}")
        return content

    @classmethod
    def to_string(cls, parsed) -> str:
        if isinstance(parsed, list):
            items = [str(item) for item in parsed]
            return "".join(items)
        else:
            return str(parsed.extract())


class HtmlTagsAttributeResolver(AbstractHtmlResolver):
    def _resolve(self, **kwargs):
        tags: List[Tag] = self.config.content

        return super()._resolve()


class ZomatoResolver(AbstractResolver):
    CACHE_EXT = 'json'
    CACHE_SUFFIX = 'zomato'
    ZOMATO_NOT_ENABLED = """
    Zomato key is not set, please get the zomato key and set add it to the configuration as property `zomato_key`.
    """

    @property
    def zomato(self) -> Pyzomato:
        return self.service.zomato

    def make_request(self) -> dict:
        return self.zomato.getDailyMenu(self.entity.selector)

    def _resolve(self, **kwargs) -> Optional[dict]:
        if self.zomato is None:
            return None
        content = self.make_request()
        log.debug(f"[ZOMATO] Response: {json.dumps(content, indent=2)}")
        return content

    def resolve_text(self, **kwargs) -> Optional[str]:
        content = self.resolve(**kwargs)
        if content is None:
            return None
        return "\n".join(self._make_lines(content))

    def _make_lines(self, content: dict) -> list:
        result = []
        menus = content['daily_menus']
        for menu in menus:
            menu = menu.get('daily_menu')
            dishes = menu['dishes']
            for dish in dishes:
                dish = dish['dish']
                result.append(f"{dish['name']} - {dish['price']}")
        return result


class PDFResolver(RequestResolver):
    CACHE_EXT = 'pdf'
    CACHE_SUFFIX = 'pdf'

    def _resolve(self, **kwargs):
        response = super()._resolve(**kwargs)
        if not response or not response.ok:
            log.error(f"Unnable to get response from: {self.url}")
            return None
        text = self._resolve_text_from_content(io.BytesIO(response.content))
        log.info(f"[PDF] Resolved text: {text}")
        return text

    def resolve_text(self, **kwargs) -> str:
        text = self.resolve(**kwargs)
        return f"PDF is available at: {self.entity.url}\n\n{text}"

    def _resolve_text_from_content(self, stream: io.BytesIO):
        out = io.BytesIO()
        laparams = pdfminer.layout.LAParams()
        for param in ("all_texts", "detect_vertical", "word_margin", "char_margin", "line_margin", "boxes_flow"):
            paramv = locals().get(param, None)
            if paramv is not None:
                setattr(laparams, param, paramv)

        high_level.extract_text_to_fp(stream, outfp=out, laparams=laparams)
        return out.getvalue().decode('utf-8')


class OcrImgRawResolver(RequestResolver):
    CACHE_EXT = 'img'
    CACHE_SUFFIX = 'img-ocr'

    def _resolve(self, **kwargs):
        response = super()._resolve(**kwargs)
        if not response or not response.ok:
            log.error(f"Unnable to get response from: {self.url}")
            return None
        text = self._resolve_text_from_content(io.BytesIO(response.content))
        log.info(f"[IMG] Resolved image: {text}")
        return text

    def resolve_text(self, **kwargs) -> str:
        text = self.resolve(**kwargs)
        return f"PDF is available at: {self.entity.url}\n\n{text}"

    def _resolve_text_from_content(self, stream: io.BytesIO):
        import pytesseract
        from PIL import Image
        img = Image.open(stream)
        return pytesseract.image_to_string(img, lang=self.entity.language)


class OCRHeavyResolver(RequestResolver):
    CACHE_EXT = 'html'
    CACHE_SUFFIX = 'html-img'

    def _parse_response(self, response: Response) -> List[Tag]:
        soap = BeautifulSoup(response.content, "lxml")
        sub = soap.select(self.entity.selector) if self.entity.selector else soap
        log.debug(f"[LUNCH] Parsed[{self.entity.name}]: {sub}")
        return sub

    def _resolve(self, **kwargs) -> Optional[str]:
        response = super()._resolve(**kwargs)
        if response is None:
            return None
        parsed = self._parse_response(response=response)
        if not parsed:
            return None
        url = parsed[0]['src']
        log.info(f"[OCR] Got an URL for [{self.entity.name}]: {url}")
        config = ResolverConfig(entity=self.entity, config=self.entity.config, content=url)

        return OcrImgRawResolver(self.service, config=config).resolve(**kwargs)

    def resolve_text(self, **kwargs) -> Optional[Any]:
        html_string = self.resolve(**kwargs)
        if html_string is None:
            return None
        return html_string


######
# Filters
######

class LunchContentFilter:
    def __init__(self, service: 'LunchService', entity: LunchEntity):
        self.entity = entity
        self.service = service

    def filter(self, content: str, **kw) -> str:
        return content


class NewLinesFilter(LunchContentFilter):
    PATTERN = re.compile("(\n+)", flags=re.MULTILINE)

    def filter(self, content, **kwargs) -> str:
        content: str = super().filter(content)
        return self.__class__.PATTERN.sub("\n", content)


class CutFilter(LunchContentFilter):
    def _find_pos(self, content, sub, diacritics=False) -> Optional[int]:
        if sub is None or content is None:
            return None
        text = unidecode.unidecode(content) if not diacritics else content
        dec_sub = unidecode.unidecode(sub) if not diacritics else sub
        log.debug(f"[CUT] Matching \"{dec_sub}\" in {text}")
        pos: re.Match = re.search(dec_sub, text, re.IGNORECASE)
        if pos is None:
            log.warning(f"[CUT] Not found position of {sub} in the content for {self.entity.name}.")
            return None
        log.info(f"[CUT] Found for {self.entity.name} suitable day delimiter for {sub} at {pos}")
        return pos.start()

    def filter(self, content: str, cut_before=None, cut_after=None, diacritics=False, **kwargs) -> Optional[str]:
        cut_before = cut_before or self.entity['cut_before']
        cut_after = cut_after or self.entity['cut_after']

        if not content:
            return None

        beg = self._find_pos(content, cut_before, diacritics=diacritics)
        end = self._find_pos(content, cut_after, diacritics=diacritics)
        if beg is None:
            beg = 0
        if end is None:
            end = len(content)
        return content[beg:end]


class DayResolveFilter(CutFilter):
    DAYS = [
        ['Pondělí', 'Úterý', 'Středa', 'Čtvrtek', 'Pátek', 'Sobota', 'Neděle'],
        ['Pondeli', 'Utery', 'Streda', 'Ctvrtek', 'Patek', 'Sobota', 'Nedele'],
        ['Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday', 'Sunday']
    ]

    @property
    def _week_day(self) -> int:
        return datetime.datetime.today().weekday()

    def options(self, day) -> Optional[List[Union[str, List[str]]]]:
        day = day
        opts = [self.entity.days] if self.entity.days is not None else []
        opts += self.__class__.DAYS
        res = [opt[day] for opt in opts if len(opt) > day]
        if not res:
            return None
        return res

    def find_day(self, day, content, diacritics=False):
        if day is None or content is None:
            return None
        day_opts = self.options(day)
        if not day_opts:
            log.warning(f"[DAY] No suitable option for a day {day} for entity {self.entity.name}")
            return None
        for opt in day_opts:
            pos = self._find_pos(content, opt, diacritics=diacritics)
            if pos is not None:
                log.info(f"[DAY] Found for {self.entity.name} suitable day delimiter for a day {day}: {opt} at {pos}")
                return pos
        return None

    def filter(self, content, day_from=None, day_to=None, diacritics=False, **kwargs):
        if not content:
            return None

        day_from = day_from or self._week_day
        day_to = day_to or (day_from + 1)
        beg = self.find_day(day_from, content, diacritics=diacritics)
        end = self.find_day(day_to, content, diacritics=diacritics)
        if beg is None:
            beg = 0
        if end is None:
            end = len(content)
        return content[beg:end]


class LunchCollection(collections.MutableMapping):
    def __init__(self, cls_wrap=None, **kwargs):
        self._collection = {key: cls_wrap(val) if cls_wrap else val for (key, val) in kwargs.items()}

    @property
    def collection(self) -> MutableMapping[str, Any]:
        return self._collection

    def __getitem__(self, k):
        return self.collection.get(k)

    def __setitem__(self, k, v):
        self.collection[k] = v

    def __delitem__(self, k):
        del self.collection[k]

    def __iter__(self):
        return iter(self.collection)

    def __len__(self):
        return len(self.collection)


class Resolvers(LunchCollection):
    def register(self, name: str, cls: type):
        if name is None or cls is None:
            return
        log.info(f"[ADD] Resolver [{name}]: {cls.__name__}")
        self._collection[name] = cls

    def get(self, name: str) -> type:
        return self._collection.get(name, HtmlResolver)

    def for_entity(self, entity: LunchEntity) -> Union[Type[ResolverChain], type]:
        if entity.resolvers is not None or entity.resolver == 'chain':
            log.info("[RESOLV] Using the chain resolver")
            return ResolverChain
        return self.get(entity.resolver)


class Filters(LunchCollection):
    def register(self, name: str, cls: type):
        log.info(f"[ADD] Filter [{name}]: {cls.__name__}")
        self._collection[name] = cls

    def get(self, name: str) -> type:
        return self._collection.get(name, LunchContentFilter)

    def for_entity(self, entity: LunchEntity) -> List[type]:
        log.debug(f"[FILTER] Filters for entity {entity.name} ~> {entity.filters}")
        return [self.get(flt) for flt in (entity.filters or [])]


class Entities(LunchCollection):
    def __init__(self, entities: dict, updated: datetime.datetime = None):
        super().__init__(cls_wrap=LunchEntity, **entities)
        self._updated = updated

    @property
    def collection(self) -> MutableMapping[str, Any]:
        if self._collection is None:
            self._collection = {}
        return self._collection

    @property
    def entities(self) -> MutableMapping[str, LunchEntity]:
        return self.collection

    def __getitem__(self, name) -> Optional[LunchEntity]:
        if name in self.entities.keys():
            instance = self.entities[name]
            log.info(f"[LUNCH] Found in entities {name}: {instance}")
            return instance
        else:
            log.warning(f"[LUNCH] Not found in entities {name}")
            return None

    def __setitem__(self, name, config):
        if name is None:
            return
        self.register(name=name, **config)

    def find_one(self, name: str) -> LunchEntity:
        return self.get(name) or self.fuz_find_one(name)[0]

    def all(self) -> ValuesView:
        return self.values()

    def find_all(self, name: str, limit=10):
        return [i[0] for i in self.fuz_find(name, limit)]

    def fuz_find(self, name: str, limit=10) -> List[Tuple]:
        return process.extract(name, self.entities,
                               processor=lambda x: x if isinstance(x, str) else x.name,
                               scorer=fuzz.token_sort_ratio,
                               limit=limit)

    def fuz_find_one(self, name: str) -> Tuple:
        return process.extractOne(name, self.entities,
                                  processor=lambda x: x if isinstance(x, str) else x.name,
                                  scorer=fuzz.token_sort_ratio)

    def register(self, name: str, url: str, display_name: str = None, tags=None,
                 selector=None, request_params=None, override=False, **kwargs):
        if name is None:
            return
        if name in self.entities.keys():
            if override:
                log.info(f"[REG] Overriding already existing: {name}")
            else:
                log.info(f"[REG] Skipping {name} since it already exists.")
                return
        config = dict(name=name, url=url, selector=selector, display_name=display_name, tags=tags,
                      request_params=request_params, **kwargs)
        instance = LunchEntity(config)
        log.info(f"[REG] Register [{name}]: {instance}")
        self.entities[name] = instance

    def all_tags(self) -> List[str]:
        accumulator = set()
        for entity in self.entities.values():
            if entity.tags:
                accumulator.update(entity.tags)
        return list(accumulator)

    def find_by_tags(self, expression: str):
        tags = TagsEvaluator(expression, self.all_tags())
        result = [entity for entity in self.entities.values() if tags.evaluate(entity.tags)]
        log.info(f"[FIND] Found by tags {expression}: {result}")
        return result

    def to_dict(self) -> dict:
        return {
            'restaurants': {name: value.config for (name, value) in self.entities.items()},
            'timestamp': self._updated.isoformat() if self._updated is not None else None
            }

    def select(self, selectors, fuzzy=False, tags=False, with_disabled=True) -> List[LunchEntity]:
        def _get() -> List['LunchEntity']:
            if selectors is None or len(selectors) == 0:
                return list(self.values())
            if tags:
                full = " ".join(selectors)
                return self.find_by_tags(full)
            return [self.find_one(select) for select in selectors]

        instances = _get()
        instances = [instance for instance in instances if instance is not None]
        if with_disabled:
            return instances

        return [item for item in instances if item and not item.disabled]

    def select_as_dict(self, selectors, tags=False, with_disabled=True) -> Mapping:
        return {item.name: item for item in self.select(selectors, tags=tags, with_disabled=with_disabled) if item}


class LunchService:
    def __init__(self, config: AppConfig, entities: Entities):
        self._entities: Entities = entities
        self._resolvers: Resolvers = Resolvers(
            default=HtmlResolver,
            zomato=ZomatoResolver,
            ocr_img=OCRHeavyResolver,
            ocr_raw=OcrImgRawResolver,
            pdf=PDFResolver,
            request=RequestResolver,
            chain=ResolverChain,
            html_tags=HtmlTagsSelectorResolver,
            html=HtmlResolver,
            html_attr=HtmlTagsAttributeResolver
        )
        self._filters: Filters = Filters(
            raw=LunchContentFilter,
            day=DayResolveFilter,
            nl=NewLinesFilter,
            cut=CutFilter
        )
        self._config: AppConfig = config
        self._zomato: Optional[Pyzomato] = None
        self._cache: LunchCache = LunchCache(self)
        self._blacklist: EntityBlacklist = EntityBlacklist(self)

    @property
    def cache(self) -> 'LunchCache':
        return self._cache

    @property
    def blacklist(self) -> 'EntityBlacklist':
        return self._blacklist

    @property
    def zomato(self) -> Optional[Pyzomato]:
        if self._zomato is None:
            if self.config.zomato_key is None:
                return None
            self._zomato = Pyzomato(self.config.zomato_key)
        return self._zomato

    @property
    def config(self) -> AppConfig:
        return self._config

    @property
    def resolvers(self) -> Resolvers:
        return self._resolvers

    @property
    def filters(self) -> Filters:
        return self._filters

    @property
    def instances(self) -> Entities:
        return self._entities

    def import_file(self, file: Tuple[Path, str], override=False):
        file = Path(file)
        if not file.exists():
            log.warning(f"[IMPORT] File not exists: {file}")
            return
        with file.open("r", encoding='utf-8') as fp:
            log.info(f"[IMPORT] Importing file: {file}")
            restaurants = yaml.safe_load(fp)
            for (name, restaurant) in restaurants['restaurants'].items():
                restaurant['name'] = name
                restaurant['override'] = override
                if restaurant.get('tags') and restaurant.get('resolver', 'default') != 'default' and restaurant[
                    'resolver'] not in restaurant['tags']:
                    restaurant['tags'].append(restaurant['resolver'])
                log.info(f"[IMP] Importing restaurant: {restaurant}")
                self.instances.register(**restaurant)

    def import_string(self, string: str, override=False):
        log.info(f"[IMPORT] Importing content: {string}")
        restaurants = yaml.safe_load(string)
        for (name, restaurant) in restaurants['restaurants'].items():
            restaurant['name'] = name
            restaurant['override'] = override
            if restaurant.get('tags') and restaurant.get('resolver', 'default') != 'default' \
                    and restaurant['resolver'] not in restaurant['tags']:
                restaurant['tags'].append(restaurant['resolver'])
            log.info(f"[IMP] Importing restaurant: {restaurant}")
            self.instances.register(**restaurant)

    def process_lunch_name(self, name: str) -> str:
        if not name or name == 'list':
            return self.to_string()
        log.info(f"[CMD] Lunch name: {name}")
        instance = self.instances.get(name)

        try:
            content = f"Restaurant: \"{name}\" - {instance.url}\n" + instance.invoke()
            log.debug(f"Content: {content}")
            return content
        except Exception as ex:
            return "ERR: {ex}"

    def __str__(self) -> str:
        return self.to_string()

    def to_string(self) -> str:
        result = f"Available ({len(self.instances)}): \n"
        for restaurant in self.instances.values():
            result += f" - {restaurant.name} - {restaurant.url}\n"
        return result

    def resolve_text(self, entity: LunchEntity, **kwargs) -> str:
        return self.cache.wrap(entity, func=self._resolve_text, ext='txt', **kwargs)

    def _resolve(self, entity, **kwargs):
        content = self._get_resolver(entity).resolve(**kwargs)

        if not content:
            log.warning(f"[SERVICE] No content for {entity.name}")
            return None
        return content

    def _resolve(self, entity, **kwargs):
        return self._cache_wrap(entity, func=self._resolve, ext='cache', **kwargs)

    def _resolve_text(self, entity: LunchEntity, **kwargs) -> Union[Optional[str], Any]:
        content = self._get_resolver(entity).resolve_text(**kwargs)
        if not content:
            log.warning(f"[SERVICE] No content for {entity.name}")
            return None

        # Do not apply filters if no filters
        if kwargs.get('no_filters'):
            return content

        return self._apply_filters(entity, content, **kwargs)

    def _get_resolver(self, entity) -> AbstractResolver:
        resolver = self.resolvers.for_entity(entity)
        log.debug(f"[RESOLVER] Using the resolver for {entity.name}: {resolver.__name__}")
        config = ResolverConfig(config=entity.config, entity=entity, content=None)

        return resolver(service=self, config=config)

    def _apply_filters(self, entity: 'LunchEntity', content: str, **kwargs):
        filters = self.filters.for_entity(entity)
        for flt in filters:
            if flt == DayResolveFilter and kwargs.get('full'):
                log.info("[FILTER] Skip the 'day' filter since full content expected.")
                continue
            log.debug(f"[FILTER] Using the text filter: {flt.__name__}")
            content = flt(self, entity).filter(content)
        return content


class EntityBlacklist:
    def __init__(self, service: LunchService):
        self._service = service

    @property
    def service(self) -> LunchService:
        return self._service

    def load(self) -> MutableMapping:
        path = self.path
        if not path.exists():
            return {}
        return json.load(self.path.open('r'))

    def save(self, blacklist: MutableMapping):
        json.dump(blacklist, self.path.open('w'))

    @property
    def path(self) -> Path:
        return self.service.cache.cache_base / self.service.cache.for_day(day=None) / 'blacklist.json'

    def get(self, entity: LunchEntity = None) -> Optional[Dict]:
        if self.service.cache.disabled:
            return None
        blacklist = self.load()
        return blacklist.get(entity.name)

    def is_blacklisted(self, entity: LunchEntity) -> bool:
        metadata = self.get(entity)
        return self.metadata_blacklisted(metadata)

    def metadata_blacklisted(self, metadata: MutableMapping) -> bool:
        if not metadata:
            return False
        now = datetime.datetime.now()
        meta_time = datetime.datetime.fromtimestamp(metadata['timestamp'])
        return meta_time > now

    def blacklist(self, entity: LunchEntity):
        name = entity.name
        if self.service.cache.disabled:
            log.info("[BLKL] Cache is disabled - not blacklisting")
            return
        blacklist = self.load()

        if blacklist is None:
            return

        if name not in blacklist.keys():
            log.info(f"[BLKL] Entity {name} not found in the blacklist!")
            base = dict(name=name, count=1)
            blacklist[name] = base
        else:
            log.info(
                f"[BLKL] Entity {name} found in the blacklist - increasing the count {blacklist[name]['count'] + 1}")
            blacklist[name]['count'] += 1

        black_time = datetime.datetime.now() + datetime.timedelta(minutes=15)
        blacklist[name]['timestamp'] = black_time.timestamp()

        self.save(blacklist)


class LunchCache:
    def __init__(self, service: 'LunchService'):
        self.service = service
        log.info(f"[CACHE] Using cache: {self.cache_base}")

    @property
    def config(self) -> AppConfig:
        return self.service.config

    @property
    def enabled(self) -> bool:
        return not self.disabled

    @property
    def disabled(self) -> bool:
        return self.config.get('no_cache', False) or self.cache_base is None

    @property
    def cache_base(self) -> Path:
        return Path(self.config.cache_dir)

    def save(self, path: Path, content):
        if self.disabled:
            log.info(f"[CACHE] Cache is not enabled or cache dir not set - not saving")
            return

        if content is None:
            log.warning(f"[CACHE] No content provided - not saving: {path}")

        log.info(f"[CACHE] Writing content to cache: {path}")
        fp: Path = self._cache_path(path)
        self._create_dir(fp.parent)
        fp.write_text(str(content), encoding='utf-8')

    def get(self, path: Path) -> Optional[str]:
        if self.disabled:
            log.debug(f"[CACHE] Cache is not enabled or cache dir not set - no content")
            return None
        fp: Path = self._cache_path(path)
        if not fp.exists():
            log.debug(f"[CACHE] Cache for item not exists: {fp}")
            return None
        return fp.read_text(encoding='utf-8')

    def _cache_path(self, fragment: Path) -> Path:
        fragment = Path(fragment)
        return self.cache_base / fragment

    def _create_dir(self, dir: Path) -> Path:
        dir = Path(dir)
        if not dir.exists():
            dir.mkdir(parents=True)
        return dir

    @property
    def _today_date(self) -> str:
        return datetime.datetime.today().strftime('%Y-%m-%d')

    def _cache(self, date=None) -> Optional[Path]:
        if self.cache_base is None:
            return None
        return Path(self._today_date if date is None else date)

    def for_day(self, day: str = None) -> Path:
        if day is None:
            day = self._today_date
        return Path(day)

    def create_fragment(self, entity: LunchEntity, day=None, suffix=None, ext='txt') -> Path:
        file_name = entity.name
        if suffix is not None:
            file_name += "-" + suffix
        return self.for_day(day) / f'{file_name}.{ext}'

    def store_entity(self, entity: LunchEntity, content: str, suffix=None, day=None, ext='txt'):
        fragment = self.create_fragment(entity, day=day, suffix=None, ext=ext)
        self.save(fragment, content)

    def get_entity(self, entity: LunchEntity, day=None, suffix=None, ext='txt'):
        if self.disabled:
            log.info("[CACHE] Cache is not enabled.")
            return None
        fragment = self.create_fragment(entity, day=day, suffix=suffix, ext=ext)
        log.info(f"[CACHE] Cache for entity {entity.name}: {fragment}")
        content = self.get(fragment)
        if not content:
            log.debug(f"[CACHE] No content for {entity.name} - {fragment}")
        return content if content else None

    def paths_for_entity(self, entity: LunchEntity, day=None, relative=False):
        if self.disabled:
            log.info("[CACHE] Cache is not enabled.")
            return None

        dir = self.for_day(day)
        fdir = self.cache_base / dir
        paths = list(fdir.glob(f"{entity.name}*"))
        if paths and relative:
            paths = [path.relative_to(self.cache_base) for path in paths]
        return paths

    def clear(self, instances=None, day=None):
        if self.disabled:
            log.info("[CACHE] Cache is not enabled.")
            return

        if instances is None:
            dir = str(self._cache_path(self.for_day(day)))
            log.info(f"[CACHE] Removing the directory: {dir}")
            shutil.rmtree(dir, ignore_errors=True)
            return [dir]

        result = []
        for inst in instances:
            files = self.paths_for_entity(inst, day=day)
            for file in files:
                result.append(str(file))
                file.unlink()
        return result

    def wrap(self, entity: LunchEntity, func, day=None, ext=None, suffix=None, **kwargs) -> str:
        if not self.enabled:
            return self._execute_func(entity=entity, func=func, **kwargs)

        cached = self.get_entity(entity, day=day, ext=ext, suffix=suffix)
        if cached:
            return cached

        content = self._execute_func(entity=entity, func=func, **kwargs)
        if content:
            self.store_entity(entity, content=content, ext=ext, suffix=suffix)
        return content

    def _execute_func(self, entity: LunchEntity, func, **kwargs):
        metadata = self.service.blacklist.get(entity)
        if metadata and self.service.blacklist.metadata_blacklisted(metadata):
            log.debug(
                f"[CACHE] Entity {entity.name} is blaclisted until {datetime.datetime.fromtimestamp(metadata['timestamp'])}.")
            return None
        result = func(entity=entity, **kwargs)
        if not result:
            log.info(f"[CACHE] Blaclisting {entity.name} for 15 minutes.")
            self.service.blacklist.blacklist(entity)
        return result

    def content(self, day=None):
        if self.disabled:
            return None

        day_path = self.for_day(day=day)
        log.info(f"[CACHE] Cache content for {day_path}: {self.cache_base / day_path}")
        full = str(self.cache_base / day_path)
        return list(os.listdir(full))


def to_text(content):
    h = html2text.HTML2Text()
    h.ignore_links = True
    h.ignore_images = True
    h.ignore_tables = True
    h.ignore_emphasis = True
    return h.handle(str(content)).strip()



import logging
from pathlib import Path
import yaml
from typing import List, Optional, Mapping, Union, MutableMapping, Any
import collections


log = logging.getLogger(__name__)


def load_yaml(file: Union[Path, str]) -> MutableMapping[str, Any]:
    file = Path(file)

    if not file.exists():
        log.warning(f"[LOAD] Config file not exists: {file}")
        return {}
    with file.open("r") as fp:
        return yaml.safe_load(fp)

def save_yaml(file: Union[Path, str], content: dict):
    file = Path(file)
    
    if not file.parent.exists():
        log.warning(f"[SAFE] Unnable to safe config file (directory not exists): {file}")
        return
    with file.open("w") as fp:
        yaml.safe_dump(content, fp)
        

def write_instances(instances, transform=None, writer=None):
    writer = writer or print
    log.debug(f"Printing: {instances}")
    if instances is None or not instances:
        writer("**No instance has been found**")

    elif isinstance(instances, list):
        for instance in instances:
            if instance is not None:
                log.info(f"Sending one response from list: {instances}")
                writer(transform(instance))
    else:
        log.info(f"Sending one response: {instances}")
        writer(transform(instances))

def generate_nice_header(*strings):
    def _for_print(max_l, curr, char='='):
        return char * (max_l - curr)

    def _print_text(max_l, text):
        buffer = ''
        buffer += f"\n===  {text}"
        buffer +=_for_print(max_l, len(text), char=' ') 
        buffer += "  ==="
        return buffer

    def _beg_end_line(max_l):
        return f"\n{_for_print(max_l + 10, 0)}"
    
    max_len = max(len(text) for text in strings)

    result = _beg_end_line(max_len)
    for text in strings:
        result += _print_text(max_len, text)
    result += _beg_end_line(max_len)
    return result + "\n\n"



class CollectionWrapper(collections.MutableMapping):
    def __init__(self, cls_wrap=None, **kwargs):
        self._collection = { key: cls_wrap(val) if cls_wrap else val for (key, val) in kwargs.items() } 

    @property
    def collection(self) -> MutableMapping[str, Any]:
        return self._collection

    def __getitem__(self, k):
         return self.collection.get(k)
    def __setitem__(self, k, v):
         self.collection[k] = v
    def __delitem__(self, k):
         del self.collection[k]
    def __iter__(self):
        return iter(self.collection)
    def __len__(self):
         return len(self.collection) 


# -*- coding: utf-8 -*-
from distutils.core import setup

packages = \
['pylunch']

package_data = \
{'': ['*'], 'pylunch': ['static/*', 'templates/*', 'templates/admin/*']}

install_requires = \
['aiodns',
 'aiohttp',
 'beautifulsoup4',
 'click',
 'flask',
 'flask-jwt-extended',
 'fuzzywuzzy',
 'gunicorn>=19.9,<20.0',
 'html2text',
 'ipython',
 'lxml',
 'pdfminer-six',
 'pillow',
 'pytesseract',
 'python-levenshtein',
 'python-telegram-bot>=12.0.0b1,<13.0.0',
 'pyyaml',
 'pyzomato',
 'requests',
 'toml>=0.10.0,<0.11.0',
 'unidecode',
 'werkzeug']

entry_points = \
{'console_scripts': ['pylunch = pylunch:cli.main_cli']}

setup_kwargs = {
    'name': 'pylunch',
    'version': '0.2.6',
    'description': 'Pylunch cli and web tool to get lunch info',
    'long_description': '# Pylunch\n\n``PyLunch`` was designed to scrape restaurants for it\'s daily menus and show them in multiple ways and formats.\n\n## Getting stated\n`PyLunch` is a scraper, to use it, you have to install it. It is available at [PyLunch](https://gitlab.com/pestanko/pylunch)\n\n### Install the released version\n\nYou can install the `PyLunch` using the `pip`:\n```bash\n$ pip install pylunch\n```\n\n\n### Install the development version from github\n- Python 3.7 or later\n- (Optional) [pipenv](https://github.com/pypa/pipenv)\n\nOr you can add it as a development dependency using the ``pipenv``\n\n```bash\n$ git clone https://github.com/pestanko/pylunch.git\n$ cd pylunch\n$ pipenv install # install the dependencies\n```\n\n### Install Tesseract for OCR support\nTo support extracting text from images you need to install [Tesseract](https://github.com/tesseract-ocr/tesseract/wiki).\nThen you need to download trained models for your language [Tesseract data files wiki](https://github.com/tesseract-ocr/tesseract/wiki/Data-Files) \n(English is supported by default).\n\nRecommended are the [best (most accurate) trained models](https://github.com/tesseract-ocr/tessdata_best).\nSave downloaded trained models to a dictionary with name `tessdata`. \nExpose environment variable `TESSDATA_PREFIX` with path to `tessdata` directory (i.e. `TESSDATA_PREFIX=/tmp/tessdata`).\n\nYou can specify different language when extracting data from the image for any restaurant with attribute `language: <Lang code>` in `restaurants.yml` file. \n\n#### Dockerfile supported languages:\n\n| Language | Lang code |\n| ---------|:---:| \n| English  | eng | \n| Czech    | ces | \n| Slovak   | slk | \n\n\n### First Run\nIn order to run the tool correctly you need to configure it.\n\nAt first run the tool will create config folder in the home directory with restaurants database, \nby default it will be empty.\n\nTo use the `zomato` resolver you need to add the `zomato_key` to your configuration.\nIn order to get the zomato key you need to go to the [Zomato Api documentation](https://developers.zomato.com/api)\n\n```bash\n$ pylunch cfg-set zomato_key "<YOUR_KEY>"\n```\n\nTo use the telegram integration, you need to set the telegram bot token property ``telegram_token``.\nTo get the telegram bot token please take a look [here](https://core.telegram.org/bots).\n\n```bash\n$ pylunch cfg-set telegram_token "<YOUR_KEY>"\n```\n\nAlso if you have already created or exported database of the restaurants, you an import it using the command `import`:\nExample file is located in: `resources/restaurants.yml` available [here](https://gitlab.com/pestanko/pylunch/raw/master/resources/restaurants.yml).\n\n```bash\n$ pylunch import -O restaurants.yml\n```\n\nIf you wish to load example from the repository you can use the `curl` and just pipe it to the `import` subcommand.\n\n```bash\n$ curl \'https://gitlab.com/pestanko/pylunch/raw/master/resources/restaurants.yml\' | pylunch import -O \n```\n\nIn order to export the restaurants database file, you can use the export command.\n```bash\n$ pylunch export -f exported.yml\n```\n\n## Run the cli tool\n\nHera is a help output with available commands.\n\n```bash\n# Show the help\n$ pylunch --help\n\nUsage: pylunch [OPTIONS] COMMAND [ARGS]...\n\n  PyLunch CLI tool\n\nOptions:\n  --version              Show the version and exit.\n  -L, --log-level TEXT   Set log level (d|i|w|e) - default=w\n  -C, --no-cache         Disable cache\n  -c, --config-dir TEXT  Location to the configuration directory\n  -F, --format TEXT      Set output format\n  --help                 Show this message and exit.\n\nCommands:\n  add            Adds a new restaurant\n  cache-clear    Clear a current cache for a day\n  cache-content  Show the current cache for a day\n  cfg-edit       Edit a configuration using the editor (Ex: VIM)\n  cfg-set        Set a config value in the user configuration\n  config         Shows the current configuration\n  console        Start the interactive console (IPython)\n  disable        Disables the restaurants\n  edit           Edits restaurants DB file\n  enable         Enables the restaurants\n  export         Export restaurants\n  import         Import restaurants\n  info           Get info for the restaurant\n  ls             List all available restaurants\n  menu           Get menu for a restaurant\n  rm             Removes the restaurant\n  telegram-bot   Start the telegram bot\n```\n\n### Runnig the server\n\nAs a server is used the Flask.\nTo run the flask server please take a look to the documnetation.\n\nIn order to start the pylunch server take a look at the `run_flask.ps1` or `run_flask.sh` scripts.\n\nFor access to the admin pages - in order to import restaurants or invalidate (refresh cache) you will need to set env variables:\n```\nPYLUNCH_SECRET=<JWT_SECRET>\nPYLUNCH_USERS=PATH_TO_USERS_FILE\n```\n\n\nAdmin user credentials:\n\n```\nUsername: admin\nPassword: foobar1\n```\n',
    'author': 'Peter Stanko',
    'author_email': 'peter.stanko0@gmail.com',
    'url': 'https://gitlab.com/pestanko/pylunch',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



def get_version():
    import toml
    from pathlib import Path
    path = Path(__file__).resolve().parents[1] / 'pyproject.toml'
    if path.exists():
        pyproject = toml.loads(open(str(path)).read())
        return pyproject['tool']['poetry']['version']
    else:
        import pkg_resources
        return pkg_resources.get_distribution("pylunch").version

__version__ = get_version()